; -------------------------------------------------------------------------
;  client-connection.scm - Ben VandenBos
;
;  *contains connection related functions and game-state related functions
; -------------------------------------------------------------------------


; Globals
; ---------
(define *you-are-index* 0)     ; Index of the client's tank in the game-state (set by server-called function "you-are")
(define *playing* #f)          ; Boolean indicating if we're ready to take movement input from user, i.e. we're connected and playing
(define *tcp-out* null)        ; TCP/IP output port (set in function "connect")
(define *tcp-in* null)         ; TCP/IP input port (set in function "connect")
(define *game-state-start-size* 20)    ; Initial game-state size (vector length)
(define *game-state* (make-vector *game-state-start-size*))  ; The game-state vector
(define *game-state-semaphore* (make-semaphore 1)) ; Semaphore for reading/writing to the game-state
(define *number-of-objects* 0) ; Number of objects in the game-state
(define *collision-thread* 0)  ; Thread descriptor for collision detecting thread


; Semaphore Declarations
; ----------------------
(define *drawing-semaphore* (make-semaphore 1))

; Debug Display Functions
; -----------------------
(define display-object
  (lambda (object-inst)
    (display (object-x object-inst))
    (display "\t")
    (display (object-y object-inst))
    (display "\t")
    (display (floor (object-theta object-inst)))
    (display "\t")
    (display (object-movement object-inst))
    (display "\t")
    (display (object-rotate object-inst))
    (display "\t")
    (display (object-time-stamp object-inst))
    (newline)))


; Client->Server Functions
; ------------------------
; used to send stuff to the server
(define send-command
  (lambda (command)
    (if *debug* (begin
                  (display "Sending: ")
                  (display command)
                  (newline)))
    (display command *tcp-out*)))

; send a chat message to the server
(define send-chat
  (lambda (str)
    (send-command `(chat ,(string-append "\"" str "\"")))))

; alert the server of a collision
(define collision
  (lambda (index)
    (send-command `(collision ,index))))

; send the fire function to the server
(define fire
  (lambda ()
    (semaphore-wait *game-state-semaphore*)
    (letrec ((pos (exact-position (vector-ref *game-state* *you-are-index*)))
             (obj-x (car pos))
             (obj-y (cadr pos))
             (obj-theta (caddr pos)))
      (semaphore-post *game-state-semaphore*)
      (send-command `(fire
                      ,*you-are-index*
                      ,obj-x
                      ,obj-y
                      ,(round-off-theta obj-theta))))))

; send the server the new state of the client's tank
(define update-server
  (lambda (index)
    (semaphore-wait *game-state-semaphore*)
    (let ((object (vector-ref *game-state* index)))
      (semaphore-post *game-state-semaphore*)
      (send-command `(update-object
                      ,index
                      ,(object-x object)
                      ,(object-y object)
                      ,(object-theta object)
                      '',(object-movement object)
                      '',(object-rotate object)
                      '',(object-color object)))
       )
    )
  )

; Remove all entries from the game-state, kill all GUI threads and the collision thread
(define clear-game-state
  (lambda ()
    (if (thread? *collision-thread*)
        (begin
          (semaphore-wait *game-state-semaphore*)
          (display "Killing collision thread... [clear-game-state]\n")
          (kill-thread *collision-thread*)
          (semaphore-post *game-state-semaphore*)))
    (semaphore-wait *game-state-semaphore*)
    (let loop ((i (- (vector-length *game-state*) 1)))
      (if (object? (vector-ref *game-state* i))
          (remove-object i))
      (if (> i 0)
          (loop (- i 1))))
    (semaphore-post *game-state-semaphore*)))


; Connection Related Functions
; ----------------------------

; opens a connection to server
(define connect
  (lambda (server port)
    (if *debug* (begin
                  (display "Attempting to connect to ")
                  (display server)
                  (display ":")
                  (display port)
                  (newline)))
    (disp-console (string-append "Attempting to connect to " server ":" port))
    (set! port (string->number port))
    (let-values (((input-port output-port)
                  (tcp-connect server port)))
      (set! *tcp-in* input-port)
      (set! *tcp-out* output-port)
      (send file-connect enable #f)
      (send file-disconnect enable #t)
      (clear-game-state)
      (let loop ((input (read *tcp-in*)))
        (if (not (eof-object? input))
            (begin
              (if *debug* (begin
                            (display "Receiving: ")
                            (display input)
                            (newline)))
              (semaphore-wait *game-state-semaphore*)
              (eval input)
              (semaphore-post *game-state-semaphore*)
              (loop (read *tcp-in*))
              )
            )
        )
      (if *debug* (display "Connection closed, leaving connection thread...\n"))
      (close-input-port *tcp-in*)
      (close-output-port *tcp-out*)
      (send file-connect enable #t)
      (send file-disconnect enable #f)
      )
    )
  )