;;; 'blue' server communication functions
;;; Justin R. Hall (c) 2003

; chat
(define chat
  (lambda (msg in out)
    (display msg)
    (newline)
    (call (tell-all-except `(chat ,(string-append (string-append "\"" msg) "\"")) #f))))
                                               
; broadcast a message to everyone except the
; exclude-me output port (returns #t when finished)
(define tell-all-except
  (lambda (cmd exclude-me)
    (let loop ((position 0)
               (len (vector-length *output-ports*))
               (out '()))
      (set! out (vector-ref *output-ports* position))
      (cond ((= position (- len 1)) #t)
            ((or (null? out)
                 (eq? out exclude-me))
             (loop (+ position 1) len out))
            (else
             (begin
               (if *debug*
                   (begin
                     (display "tell-all-except: ") 
                     (display cmd)
                     (newline)))
               (display cmd out)
               (loop (+ position 1) len out)))))))

; sends the game-state to 'out'
(define send-game-state
  (lambda (game-state out)
    (let loop ((position 0)
               (len (vector-length game-state))
               (obj '()))
      (set! obj (vector-ref game-state position))
      (cond ((= position (- len 1)) #t)
            ((null? obj) (loop (+ position 1) len obj))
            (else
             (begin
               (let ((params
                      `(add-object
                        ,position
                        ,(object-type obj)
                        ,(object-x obj)
                        ,(object-y obj)
                        ,(object-theta obj)
                        ,(object-movement obj)
                        ,(object-rotate obj)
                        ,(object-color obj)
                        ,(object-owner obj))))
                 (if *debug*
                     (begin
                       (display "sending from *game-state*: ")
                       (display params)
                       (newline)))
                 (display params out))
               (loop (+ position 1) len obj)))))))

; closes down an open TCP connection
(define close-connection
  (lambda (in out)
    ; need a way to remove any/all objects associated with this connection
    (call (remove-assoc-tank *game-state* out))

    ; remove it from *output-ports* and remove reference
    ; to the thread descriptor
    (let ((index (vector-find *output-ports* out eq?)))
      (if index
          (begin
            ; clean-up
            (vector-set! *output-ports* index '())
            (vector-set! *threads* index '())
            (hash-table-remove! *ports-assoc* index)
            (hash-table-remove! *obj-assoc* out)
            
            ; decrement connection-index
            ;(set! *connection-index* (vector-num-elements *output-ports*))
            (set! *connection-index* (- *connection-index* 1))
            (if (< *connection-index* 0) (set! *connection-index* 0))
            
            ; (no need to kill thread, it will terminate after this is done)
            ; shutdown the ports
            (close-output-port out)
            (close-input-port in)
            (if *debug*
                (begin
                  (display "Connection terminated.")
                  (newline))))
          (begin
            (display "warning: attempted to remove non-existent output-port")
            (newline))))))

;;; generic multi-threaded TCP server (accepting N connections)
(define server
  (lambda (port)
    (let ((listener (tcp-listen port)))
      (let loop ()
        (if (tcp-accept-ready? listener)
          (let-values (((client->server server->client)
                        (tcp-accept listener)))

            (if *debug*
                (begin
                  (display "TCP Connection accepted...")
                  (newline)))

            (set! *connection-index* (next-available-index *output-ports*))
            
            ; pass connection off to its own thread of execution
            (let ((current-size (vector-length *threads*)))
              (if (>= *connection-index* current-size)
                  (begin
                    (set! *threads* (vec-grow *threads* (* 2 current-size)))
                    (set! *output-ports* (vec-grow *output-ports* (* 2 current-size)))
                    (vector-set! *threads* *connection-index*
                                 (thread (lambda () (handle-client client->server 
                                                                   server->client))))
                    (vector-set! *output-ports* *connection-index*
                                 server->client))
                  (begin
                    
                    (vector-set! *threads* *connection-index*
                                 (thread (lambda () (handle-client client->server 
                                                                   server->client))))
                    (vector-set! *output-ports* *connection-index*
                                 server->client))))
            
            (if *debug*
                (begin
                  (display "Current value of *connection-index*: ")
                  (display *connection-index*)
                  (newline)
                  (display "*output-ports*: ")
                  (display *output-ports*)
                  (newline)
                  (display "*threads*: ")
                  (display *threads*)
                  (newline)))
            
            (loop))
          (loop))))))

;;; manage the ports and pass them to the appropriate handler
(define handle-client
  (lambda (client->server server->client)
    ; keep track of the output port for broadcast messaging
    (cons server->client *output-ports*)
    ; call the actual client handler
    (blue-server client->server server->client)
    ; shutdown the connection if not already closed
    (close-output-port server->client)
    (close-input-port client->server)))

;;; the actual server for 'blue'
(define blue-server
  (lambda (in out)
    (begin
      ; handshake to make sure versions are the same
      (if (not (handshake in out))
          (begin (close-connection in out) #f))
      
      (let loop ((input (read in)))
        ; debugging
        (if *debug* 
            (begin
              (display "Received: ") 
              (display input) 
              (newline)))
        
        ; check for EOF, otherwise EVAL the input
        ; and tell everybody else about it
        (cond
          ((eof-object? input)
           (close-connection in out))
          ((valid-command? input)
           (begin
             (if *debug*
                 (begin
                   (display "Valid command received: ") 
		   (display (car input)) 
 		   (newline)))
             (set! input (append input `(,in) `(,out)))
             (thread (lambda () (call (eval input))))
             (loop (read in))))
          ((not (valid-command? input))
           (begin
             (display "Invalid command received: ") (display (car input)) (newline)))
          (else (loop (read in))))))))
