;;; 'blue' server object functions
;;; Justin R. Hall (c) 2003

; removes the tank associated with a particular output-port
; from the game-state
(define remove-assoc-tank
  (lambda (game-state out)
    (let ((success #t))
      (begin
        (set! success (hash-table-get *obj-assoc* out (lambda () (#f))))
        (if success
            (begin
              (vector-set! game-state success '())
              (call (tell-all-except `(remove-object ,success) out)))
            (display "error: could not find tank to remove"))))))

; removes a bullet from the game-state and tells everyone about it
(define remove-bullet
  (lambda (game-state index)
    (let ((obj (vector-ref game-state index)))
      (if (object? obj)
          (let ((td (object-thread-desc obj)))
            (if (not (eqv? (current-thread) td))
                (kill-thread td))
            (vector-set! game-state index '())
            (call (tell-all-except `(remove-object ,index) #f)))))))
  
; calculates the next available index into the game-state
(define next-available-index
  (lambda (game-state)
    (let loop ((i 0))
      (cond ((>= i (vector-length game-state)) i)
            ((null? (vector-ref game-state i)) i)
            (else (loop (+ i 1)))))))

; add a bullet object and tell everybody about it
(define fire
  (lambda (index x y theta in out)
    (let* ((next-index (call (next-available-index *game-state*)))
           (add-bullet `(add-object
                         ,next-index
                         'bullet
                         ,x
                         ,y
                         ,theta
                         'up
                         'stop
                         'black
                         ,index)))
      
      (call (add-object `(,next-index 'bullet ,x ,y ,theta 'up 'stop 'black ,index ,in ,out)))
      (call (tell-all-except add-bullet #f)))))

; adds an object to the master game-state
(define add-object
  (lambda (params)
    (let* ((v-params (list->vector params))
           (index      (vector-ref v-params 0))
           (type       (vector-ref v-params 1))
           (x          (vector-ref v-params 2))
           (y          (vector-ref v-params 3))
           (theta      (vector-ref v-params 4))
           (movement   (vector-ref v-params 5))
           (rotate     (vector-ref v-params 6))
           (color      (vector-ref v-params 7))
           (owner      (vector-ref v-params 8))
           (in         (vector-ref v-params 9))
           (out        (vector-ref v-params 10))
           (current-size (vector-length *game-state*))
           (td '()))
      
      (if *debug*
          (begin
            (display "add-object params: ")
            (display params)
            (newline)))            
      
      (begin
        ; check current size of the *game-state*
        (if (>= index current-size)
            ; resize the *game-state* vector and call add-object again
            (begin
              (set! *game-state* (call (vec-grow *game-state* (* 2 current-size))))
              (call (add-object params)))
            
            ; there's enough room, so add the object
            (begin
              ; put the new tanks *game-state* index into a hash table to
              ; associate it with its output-port
              (if (eqv? (eval type) 'tank)
                  (begin
                    (hash-table-put! *ports-assoc* index `(,in ,out))
                    (hash-table-put! *obj-assoc* out index)))

              ; put a time-to-live on the bullet
              (if (eqv? (eval type) 'bullet)
                  (set! td (thread (lambda () (sleep *bullet-life*) (call (remove-bullet *game-state* index))))))
              
              ; check for proper types
              (if (or (eqv? (eval type) 'tank)
                      (eqv? (eval type) 'bullet))
                  (vector-set! *game-state* index
                               (make-object type x y theta movement rotate color owner 0 td))
                  #f)))))))

; update an object currently in the master game-state
(define update-object
  (lambda (index x y theta movement rotate color in out)
    (if (or (>= index (vector-length *game-state*))
            (null? (vector-ref *game-state* index)))
        (begin
          (display "error: attempted to update invalid object index ")
          (display index)
          (display " (out of bounds or non-existent)")
          (newline)
          #f)
        (begin
          (set-object-x! (vector-ref *game-state* index) x)
          (set-object-y! (vector-ref *game-state* index) y)
          (set-object-theta! (vector-ref *game-state* index) theta)
          (set-object-movement! (vector-ref *game-state* index) movement)
          (set-object-rotate! (vector-ref *game-state* index) rotate)
          (set-object-color! (vector-ref *game-state* index) color)
          
          (if *debug*
              (begin
                (display "Updating object locally: ")
                (display (eval (vector-ref *game-state* index)))
                (newline)))
          
          ; tell everybody else
          (let* ((obj (vector-ref *game-state* index))
                 (cmd
                  `(update-object
                    ,index
                    ,(object-x obj)
                    ,(object-y obj)
                    ,(object-theta obj)
                    ,(object-movement obj)
                    ,(object-rotate obj)
                    ,(object-color obj))))
            (call (tell-all-except cmd out)))))))

